/* $Id: buddy.c,v 1.9 2001/11/04 01:07:03 robey Exp $
 *
 * Manipulate the buddy table (a hash mapping user id's to the list of
 * online clients who should receive notifications on them) and do some
 * standard buddy-list operations.
 *
 * Copyright (C) 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include "client.h"
#include "client-db.h"
#include "buddy.h"

/* id => GList<Client *> */
/* for each id, a list of online clients that are subscribed and not blocked */
static GHashTable *buddy_table = NULL;



void
buddy_table_init (void)
{
        buddy_table = g_hash_table_new(g_str_hash, g_str_equal);
}

static gboolean
buddy_foreach_remove (char *key, GList *value)
{
        g_free(key);
        g_list_free(value);
        return TRUE;
}

void
buddy_table_free (void)
{
        g_hash_table_foreach_remove(buddy_table, (GHRFunc)buddy_foreach_remove, NULL);
}

/* this client has just changed status.  update all buddies. */
void
buddy_update_status (Client *client, gboolean online)
{
        char *id_lower;
        GList *buddy_list, *iter;
        Client *buddy;
        EnamlBlock *status_block;

        if (client->flags & CLIENT_GUEST) {
                return;
        }

        id_lower = g_strdup(client->id);
        g_strdown(id_lower);
        buddy_list = (GList *)g_hash_table_lookup(buddy_table, id_lower);
        g_free(id_lower);
        if (buddy_list == NULL) {
                return;
        }

        status_block = client_make_status(client, online);
        for (iter = g_list_first(buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Client *)iter->data;
                client_send(buddy, status_block);
        }
        enaml_free(status_block);
}

/* new buddy for an online user */
static void
buddy_table_add_buddy (const Client *client, const char *id)
{
        char *lower_id, *orig_key;
        GList *buddy_list;
        Client *buddy_client;
        EnamlBlock *block;

        lower_id = g_strdup(id);
        g_strdown(lower_id);
        buddy_list = NULL;
        if (g_hash_table_lookup_extended(buddy_table, lower_id, (void **)&orig_key,
                                         (void **)&buddy_list)) {
                g_hash_table_remove(buddy_table, orig_key);
                g_free(orig_key);
        }
        buddy_list = g_list_prepend(buddy_list, (void *)client);
        g_hash_table_insert(buddy_table, lower_id, buddy_list);

        /* is the user on right now? */
        buddy_client = client_by_id(lower_id, CLIENT_USER);
        if (buddy_client != NULL) {
                block = client_make_status(buddy_client, TRUE);
                client_send(client, block);
                enaml_free(block);
        }
}

static void
buddy_table_delete_buddy (const Client *client, const char *id)
{
        char *lower_id, *orig_key;
        GList *buddy_list;

        lower_id = g_strdup(id);
        g_strdown(lower_id);
        buddy_list = NULL;
        if (g_hash_table_lookup_extended(buddy_table, lower_id, (void **)&orig_key,
                                         (void **)&buddy_list)) {
                g_hash_table_remove(buddy_table, orig_key);
                g_free(orig_key);
        }
        buddy_list = g_list_remove(buddy_list, (void *)client);
        if (buddy_list != NULL) {
                g_hash_table_insert(buddy_table, lower_id, buddy_list);
        } else {
                g_free(lower_id);
        }
}

/* after a client has logged in, need to add their buddy list to the tables */
void
buddy_list_load (Client *client)
{
        Buddy *buddy;
        GList *iter;

        for (iter = g_list_first(client->buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if ((buddy->flags & (BUDDY_APPROVED|BUDDY_BLOCKED)) == BUDDY_APPROVED) {
                        buddy_table_add_buddy(client, buddy->id);
                }
        }
}

/* likewise, when the client logs out, remove their buddy list from the tables */
void
buddy_list_unload (Client *client)
{
        Buddy *buddy;
        GList *iter;

        for (iter = g_list_first(client->buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if ((buddy->flags & (BUDDY_APPROVED|BUDDY_BLOCKED)) == BUDDY_APPROVED) {
                        buddy_table_delete_buddy(client, buddy->id);
                }
        }
}

/* go through a client's buddy list and remind them of any pending requests */
void
buddy_list_refresh_pending (void *db_state, Client *client)
{
        GList *iter;
        Buddy *buddy;

        for (iter = g_list_first(client->reverse_buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if (!(buddy->flags & (BUDDY_APPROVED|BUDDY_DENIED))) {
                        client_sendf(client, "subscription-request", "id:%s to-id:%s text:%s",
                                     buddy->id, client->id, buddy->message);
                }
        }

        for (iter = g_list_first(client->buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if (buddy->flags & BUDDY_STORED) {
                        /* stored approved/denied message */
                        if (buddy->flags & BUDDY_APPROVED) {
                                client_sendf(client, "subscription-accepted", "id:%s text:%s",
                                             buddy->id, buddy->message);
                        } else if (buddy->flags & BUDDY_DENIED) {
                                client_sendf(client, "subscription-denied", "id:%s text:%s",
                                             buddy->id, buddy->message);
                        }
                        buddy->flags &= ~BUDDY_STORED;
                        g_free(buddy->message);
                        buddy->message = NULL;
                        
                        if (! db_update_buddy(db_state, client, buddy)) {
                                g_warning("buddy update failed :(");
                        }
                }
        }
}

/* tell a client about everyone subscribed to her */
static void
buddy_list_refresh_stuff (Client *client, const char *tag, GList *buddy_list,
			  gboolean show_blocked)
{
        GList *iter;
        EnamlList *approved_list, *blocked_list, *pending_list;
        Buddy *buddy;

        approved_list = enaml_new_list();
        pending_list = enaml_new_list();
	if (show_blocked) {
		blocked_list = enaml_new_list();
	}
        for (iter = g_list_first(buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if (buddy->flags & BUDDY_APPROVED) {
                        if (show_blocked && (buddy->flags & BUDDY_BLOCKED)) {
                                enaml_list_add_string(blocked_list, buddy->id, 0);
                        } else {
                                enaml_list_add_string(approved_list, buddy->id, 0);
                        }
                } else if (!(buddy->flags & BUDDY_DENIED)) {
                        enaml_list_add_string(pending_list, buddy->id, 0);
                }
        }

	if (show_blocked) {
		client_sendf(client, tag, "approved:%L blocked:%L pending:%L",
			     approved_list, blocked_list, pending_list);
	} else {
		client_sendf(client, tag, "approved:%L pending:%L",
			     approved_list, pending_list);
	}
}

void
buddy_list_refresh (Client *client)
{
        buddy_list_refresh_stuff(client, "subscriptions", client->buddy_list, FALSE);
}

void
buddy_list_refresh_reverse (Client *client)
{
        buddy_list_refresh_stuff(client, "subscriptions-from", client->reverse_buddy_list, TRUE);
}

void
buddy_list_handle_unsubscribe (void *db_state, Client *client, Socket *sock, const char *id)
{
        Buddy *buddy;
        Client *client_r;

        buddy = client_buddy_find(client, id);
        if ((buddy == NULL) || (buddy->flags & BUDDY_DENIED)) {
                send_enaml_buildf(sock, "no-subscription", "id:%s", id);
                return;
        }

        if (! db_delete_buddy(db_state, client, buddy)) {
                g_warning("buddy delete failed :(");
                send_enaml_buildf(sock, "error", "text:%s", "Database operation failed!");
                return;
        }

        client_r = client_by_id(buddy->id, CLIENT_USER);
        if (client_r != NULL) {
                /* ex-buddy is online: tell them of the change */
                client_sendf(client_r, "unsubscribed-from", "id:%s", client->id);
                client_buddy_remove_reverse(client_r, client->id);
                /* send "signoff" notice for this online user */
                client_sendf(client, "status", "nick:%s id:%s time:%d mode:%s",
                             client_r->nick, client_r->id, (long)time(NULL), "offline");
        }

        /* remove buddy from internal tables */
        client_buddy_remove(client, buddy->id);
        buddy_table_delete_buddy(client, buddy->id);

        /* refresh buddy list info to client */
        buddy_list_refresh(client);
}
                
void
buddy_list_handle_subscribe (void *db_state, Client *client, Socket *sock, const char *id, const char *text)
{
        Buddy *buddy, *buddy_r;
        Client *client_r;

        buddy = client_buddy_find(client, id);
        if (buddy != NULL) {
                if (buddy->flags & BUDDY_APPROVED) {
                        client_sendf(client, "subscription-accepted", "id:%s already", id);
                        return;
                }
                if (buddy->flags == 0) {
                        /* pending: oh be quiet */
                        return;
                }
                if (buddy->deny_count >= CLIENT_MAX_BUDDY_DENIALS) {
                        client_sendf(client, "subscription-denied", "id:%s patience-exceeded", id);
                        return;
                }

                /* previously denied, but try again */
                buddy->flags = 0;
		g_free(buddy->message);
		buddy->message = g_strndup(text, CLIENT_MAX_BUDDY_MESSAGE);

                client_r = client_by_id(buddy->id, CLIENT_USER);
                if (client_r != NULL) {
			/* target is online: update the reverse mapping too */
                        buddy_r = client_buddy_find_reverse(client_r, client->id);
                        buddy_r->flags = 0;
                        g_free(buddy_r->message);
                        buddy_r->message = g_strndup(text, CLIENT_MAX_BUDDY_MESSAGE);
                }

                if (! db_update_buddy(db_state, client, buddy)) {
                        g_warning("buddy update failed :(");
                        send_enaml_buildf(sock, "error", "text:%s", "Database operation failed!");
                        return;
                }
        } else {
		/* new buddy request */
                buddy = client_buddy_add(client, id, 0, text, 0);

                client_r = client_by_id(id, CLIENT_USER);
                if (client_r != NULL) {
			/* target is online: create a new reverse mapping too */
			client_buddy_add_reverse(client_r, client->id, 0, text);
                }

		if (! db_add_buddy(db_state, client, buddy)) {
                        g_warning("buddy add failed :(");
                        send_enaml_buildf(sock, "error", "text:%s", "Database operation failed!");
                        return;
                }
        }

	/* finally, send the request message */
        if (client_r != NULL) {
                client_sendf(client_r, "subscription-request", "id:%s to-id:%s text:%s",
                             client->id, client_r->id, text);
        }
}

void
buddy_list_handle_accept (void *db_state, Client *client, Socket *sock, const char *id, const char *text)
{
        Buddy *buddy, *buddy_r;
        Client *client_r;

        buddy = client_buddy_find_reverse(client, id);
        if ((buddy == NULL) || (buddy->flags & BUDDY_DENIED)) {
                send_enaml_buildf(sock, "no-subscription-request", "id:%s", id);
                return;
        }

        if (buddy->flags & BUDDY_APPROVED) {
                /* already approved */
                buddy_list_refresh_reverse(client);
                return;
        }
        buddy->flags = BUDDY_APPROVED;
        g_free(buddy->message);
        buddy->message = g_strdup(text);

        client_r = client_by_id(buddy->id, CLIENT_USER);
        if (client_r != NULL) {
                /* origin is online; give them the good news */
                client_sendf(client_r, "subscription-accepted", "id:%s text:%s", client->id, text);
                buddy_table_add_buddy(client_r, client->id);
        } else {
                client_r = client_lookup_offline(db_state, buddy->id, CLIENT_USER);
                g_assert(client_r != NULL);
                db_get_buddies(db_state, client_r);
        }

        buddy_r = client_buddy_find(client_r, client->id);
        g_assert(buddy_r != NULL);
        buddy_r->flags = BUDDY_APPROVED;
        buddy_r->deny_count = 0;
        g_free(buddy_r->message);
        if (client_r->flags & CLIENT_OFFLINE) {
                /* store the approved message for later */
                buddy_r->message = g_strdup(text);
                buddy_r->flags |= BUDDY_STORED;
        } else {
                buddy_r->message = NULL;
        }

        if (! db_update_buddy(db_state, client_r, buddy_r)) {
                g_warning("buddy update failed :(");
                send_enaml_buildf(sock, "error", "text:%s", "Database operation failed!");
                return;
        }
        buddy_list_refresh_reverse(client);
        client_discard(client_r);
}

void
buddy_list_handle_denied (void *db_state, Client *client, Socket *sock, const char *id, const char *text)
{
        Buddy *buddy, *buddy_r;
        Client *client_r;
        gboolean revoke = FALSE;

        buddy = client_buddy_find_reverse(client, id);
        if ((buddy == NULL) || (buddy->flags & BUDDY_DENIED)) {
                send_enaml_buildf(sock, "no-subscription-request", "id:%s", id);
                return;
        }

        if (buddy->flags & BUDDY_APPROVED) {
                /* okay for flags to be BUDDY_APPROVED here: the user would like to revoke a buddy, basically. */
                revoke = TRUE;
        }

        buddy->flags = BUDDY_DENIED;
        g_free(buddy->message);
        buddy->message = g_strdup(text);

        client_r = client_by_id(buddy->id, CLIENT_USER);
        if (client_r != NULL) {
                /* origin is online; give them the bad news */
                client_sendf(client_r, "subscription-denied", "id:%s text:%s", client->id, text);
                if (revoke) {
                        buddy_table_delete_buddy(client_r, client->id);
                        /* send "signoff" notice for this online user */
                        client_sendf(client_r, "status", "nick:%s id:%s time:%d mode:%s",
                                     client->nick, client->id, (long)time(NULL), "offline");
                }
        } else {
                client_r = client_lookup_offline(db_state, buddy->id, CLIENT_USER);
                g_assert(client_r != NULL);
                db_get_buddies(db_state, client_r);
        }
        
        buddy_r = client_buddy_find(client_r, client->id);
        g_assert(buddy_r != NULL);
        buddy_r->flags = BUDDY_DENIED;
        buddy_r->deny_count++;
        g_free(buddy_r->message);
        if (client_r->flags & CLIENT_OFFLINE) {
                /* store the denied message for later */
                buddy_r->flags |= BUDDY_STORED;
                buddy_r->message = g_strdup(text);
        } else {
                buddy_r->message = NULL;
        }

        if (! db_update_buddy(db_state, client_r, buddy_r)) {
                g_warning("buddy update failed :(");
                send_enaml_buildf(sock, "error", "text:%s", "Database operation failed!");
                return;
        }
        buddy_list_refresh_reverse(client);
        client_discard(client_r);
}

static void
buddy_list_dump_foreach (const char *id, GList *buddy_list, GList **list)
{
        GString *str;
        Client *client;
        GList *iter;

        str = g_string_new("");
        g_string_sprintfa(str, "%s -> ", id);
        for (iter = g_list_first(buddy_list); iter != NULL; iter = g_list_next(iter)) {
                client = (Client *)iter->data;
                g_string_sprintfa(str, "%s ", client->id);
        }
        *list = g_list_prepend(*list, str->str);
        g_string_free(str, FALSE);
}

/* for debugging */
GList *
buddy_table_dump (void)
{
        GList *list = NULL;

        g_hash_table_foreach(buddy_table, (GHFunc)buddy_list_dump_foreach, &list);
        list = g_list_reverse(list);
        return list;
}

void
buddy_table_dump_free (GList *list)
{
        g_list_foreach(list, (GFunc)g_free, NULL);
        g_list_free(list);
}
