/* $Id: client-gdbm.c,v 1.7 2001/09/02 20:06:24 robey Exp $
 *
 * Do gdbm queries to fetch/create users/channels, filling in Client entries.
 *
 * gdbm is quite a bit more primitive than SQL (which isn't exactly a pillar
 * of modernity itself) so we set our sights really low here.  Basically we
 * convert the client data to a string when writing it to the database, and
 * convert it back when reading.  We use two database files: one for users,
 * and one for channels.
 *
 * For stored messages, we use the destination id as the key, and the messages
 * (since we know they're enaml blocks) are clumped into one data block,
 * separated by '\n'.
 *
 * For buddy lists, each user has two entries:
 * ">id" is a comma-separated list of id's buddies.
 * "<id" is a comma-separated list of people subscribed to id.
 * "id/buddy" is an enaml block of info about that pair.
 * And we just pray that the lists never get out of sync.
 *
 * Copyright (C) 2000, 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <say2/say2.h>
#ifdef HAVE_GDBM
#include <gdbm.h>
#else
#include <ndbm.h>
#endif
#ifndef GDBM_SYNC
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include "client.h"
#include "client-db.h"
#include "client-gdbm.h"

#define BLOCK_SIZE	4096

#ifndef HAVE_GDBM
/* alias standard gdbm stuff into ndbm equivalents */
#define GDBM_FILE DBM *
#define gdbm_open(filename, blocksize, flags, mode, fatalfunc) \
	dbm_open(filename, flags, mode)
#define gdbm_fetch dbm_fetch
#define gdbm_store dbm_store
#define gdbm_close dbm_close
#define GDBM_INSERT DBM_INSERT
#define GDBM_REPLACE DBM_REPLACE
#define GDBM_WRITER O_RDWR|O_CREAT
#endif	/* HAVE_GDBM */

#ifndef GDBM_SYNC
/* GDBM 1.7 (not everyone's got 1.8, robey) doesn't have this */
#define GDBM_SYNC O_SYNC
#endif /* GDBM_SYNC */

typedef struct {
        GDBM_FILE userdb;
        GDBM_FILE chandb;
	GDBM_FILE msgdb;
        GDBM_FILE buddydb;
} GDBMInfo;

/* convert client to an enaml string */
static void
client_update_block (const Client *client, EnamlBlock *block)
{
        char *prefs, *privs;

        enaml_add_string (block, "id", client->id, 0);
        enaml_add_string (block, "email", client->email, 0);
        if (client->tagline != NULL) {
                enaml_add_string (block, "tagline", client->tagline, 0);
        } else {
                enaml_drop_key (block, "tagline");
        }
        prefs = client_prefs_to_string (client);
        enaml_add_string (block, "prefs", prefs, ENAML_COPY);
        g_free (prefs);
        if (client->icon != NULL) {
                enaml_add_blob (block, "icon", client->icon, client->iconsize, 0);
        } else {
                enaml_drop_key (block, "icon");
        }
        if (client->attributes != NULL) {
                enaml_add_block (block, "attributes", client->attributes, 0);
        } else {
                enaml_drop_key (block, "attributes");
        }

        if (client->flags & CLIENT_USER) {
                privs = client_privs_to_string (client);
                enaml_add_string (block, "privs", privs, ENAML_COPY);
                g_free (privs);
        }
}

static EnamlBlock *
read_id_block (GDBMInfo *info, const char *id, int flags)
{
        EnamlBlock *block;
        datum key, value;
        char *id_lower;

        id_lower = g_strdup (id);
        g_strdown (id_lower);
        key.dptr = id_lower;
        key.dsize = strlen (id_lower)+1;
        if (flags & CLIENT_USER) {
                value = gdbm_fetch (info->userdb, key);
        } else {
                value = gdbm_fetch (info->chandb, key);
        }
        g_free (id_lower);

        if (value.dptr == NULL) {
                return NULL;
        }
        block = enaml_parse (value.dptr);
        free (value.dptr);
        return block;
}

static EnamlBlock *
read_client_block (GDBMInfo *info, const Client *client)
{
        return read_id_block (info, client->id, client->flags);
}

static gboolean
write_client_block (GDBMInfo *info, const Client *client, EnamlBlock *block, gboolean exclusive)
{
        char *str;
        char *id_lower;
        datum key, value;
        int ret;

        str = enaml_compose (block, FALSE);
        id_lower = g_strdup (client->id);
        g_strdown (id_lower);
        key.dptr = id_lower;
        key.dsize = strlen (id_lower)+1;
        value.dptr = str;
        value.dsize = strlen (str)+1;

        if (client->flags & CLIENT_CHANNEL) {
                ret = gdbm_store (info->chandb, key, value, exclusive ? GDBM_INSERT : GDBM_REPLACE);
        } else {
                ret = gdbm_store (info->userdb, key, value, exclusive ? GDBM_INSERT : GDBM_REPLACE);
        }

        g_free (str);
        g_free (id_lower);
        return (ret == 0);
}

#ifdef HAVE_GDBM
static void
client_gdbm_fatal (const char *errmsg)
{
        log ("ERROR: gdbm: %s", errmsg);
}
#endif

/* ---------- */

static void *
client_gdbm_init (const GData *params)
{
        const char *filename;
        char *filename_user, *filename_chan, *filename_msg, *filename_buddy;
        GDBMInfo *info;

        filename = (const char *)g_datalist_get_data((GData **)&params, "filename");
#ifdef HAVE_GDBM
        filename_user = g_strdup_printf("%s-user.db", filename);
        filename_chan = g_strdup_printf("%s-chan.db", filename);
	filename_msg = g_strdup_printf("%s-msg.db", filename);
        filename_buddy = g_strdup_printf("%s-buddy.db", filename);
#else
	filename_user = g_strdup_printf("%s-user", filename);
	filename_chan = g_strdup_printf("%s-chan", filename);
	filename_msg = g_strdup_printf("%s-msg", filename);
        filename_buddy = g_strdup_printf("%s-buddy", filename);
#endif

        info = g_new0(GDBMInfo, 1);
        info->userdb = gdbm_open(filename_user, BLOCK_SIZE, GDBM_WRITER | GDBM_SYNC, 0600, &client_gdbm_fatal);
        info->chandb = gdbm_open(filename_chan, BLOCK_SIZE, GDBM_WRITER | GDBM_SYNC, 0600, &client_gdbm_fatal);
        info->msgdb = gdbm_open(filename_msg, BLOCK_SIZE, GDBM_WRITER | GDBM_SYNC, 0600, &client_gdbm_fatal);
        info->buddydb = gdbm_open(filename_buddy, BLOCK_SIZE, GDBM_WRITER | GDBM_SYNC, 0600, &client_gdbm_fatal);
        g_free(filename_user);
        g_free(filename_chan);
	g_free(filename_msg);
        g_free(filename_buddy);

        if ((info->userdb == NULL) || (info->chandb == NULL) || (info->msgdb == NULL) || (info->buddydb == NULL)) {
                g_free(info);
                return NULL;
        }
        return info;
}

static void
client_gdbm_shutdown (void *blob)
{
        GDBMInfo *info = (GDBMInfo *)blob;

        gdbm_close(info->userdb);
        gdbm_close(info->chandb);
	gdbm_close(info->msgdb);
        gdbm_close(info->buddydb);
        g_free(info);
}

static gboolean
client_gdbm_add_client (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = enaml_new ();
        client_update_block (client, block);
        enaml_add_long (block, "created", (long) time (NULL));
        ret = write_client_block (info, client, block, TRUE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_update_login (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }

        enaml_add_long (block, "last-login", (long) time (NULL));
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_update_logout (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }

        enaml_add_long (block, "last-logout", (long) time (NULL));
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_update_client (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }

        client_update_block (client, block);
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_lookup_client (void *blob, const char *id, Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        const EnamlBlock *attr_block;
        const char *icon;

        if (client == NULL) {
                /* only wanna know if it exists */
                block = read_id_block (info, id, CLIENT_USER);
                if (block == NULL) {
                        return FALSE;
                }
                enaml_free (block);
                return TRUE;
        }

        block = read_id_block (info, id, client->flags);
        if (block == NULL) {
                return FALSE;
        }

        g_free (client->id);
        client->id = g_strdup (enaml_get_string (block, "id", NULL));
        g_free (client->nick);
        client->nick = g_strdup (client->id);
        g_free (client->email);
        client->email = g_strdup (enaml_get_string (block, "email", NULL));
        g_free (client->tagline);
        client->tagline = g_strdup (enaml_get_string (block, "tagline", NULL));
        client_string_to_prefs (client, enaml_get_string (block, "prefs", NULL));
        if (client->flags & CLIENT_USER) {
                client_string_to_privs (client, enaml_get_string (block, "privs", NULL));
        }

        client->iconsize = enaml_get_blob (block, "icon", &icon);
        if (icon != NULL) {
                client->icon = g_malloc (client->iconsize);
                memcpy (client->icon, icon, client->iconsize);
        }
        attr_block = enaml_get_block (block, "attributes");
        if (attr_block != NULL) {
                client->attributes = enaml_copy_block (attr_block);
        }

        enaml_free (block);

        return TRUE;
}

static gboolean
client_gdbm_check_password (void *blob, const char *id, const char *password, int flags)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        const char *c_password;
        gboolean ok;

        block = read_id_block (info, id, flags);
        if (block == NULL) {
                return FALSE;
        }
        c_password = enaml_get_string (block, "password", NULL);
        ok = FALSE;
        if (c_password != NULL) {
                if (strcmp (c_password, password) == 0) {
                        ok = TRUE;
                }
        }

        enaml_free (block);
        return ok;
}

static gboolean
client_gdbm_set_password (void *blob, const Client *client, const char *password)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }
        enaml_add_string (block, "password", password, 0);
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_get_seen (void *blob, const char *id, time_t *last_login, time_t *last_logout, int flags)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;

        block = read_id_block (info, id, flags);
        if (block == NULL) {
                return FALSE;
        }
        *last_login = (time_t) enaml_get_long (block, "last-login", 0L);
        *last_logout = (time_t) enaml_get_long (block, "last-logout", 0L);
        enaml_free (block);
        return TRUE;
}

static gboolean
client_gdbm_get_messages (void *blob, const Client *client, GList **messages)
{
	GDBMInfo *info = (GDBMInfo *)blob;
	datum key, value;
	char *id_lower;
	char **strv;
	int i;

	id_lower = g_strdup(client->id);
	g_strdown(id_lower);
	key.dptr = id_lower;
	key.dsize = strlen(id_lower)+1;
	value = gdbm_fetch(info->msgdb, key);

	if (value.dptr == NULL) {
		g_free(id_lower);
		return FALSE;
	}

	strv = g_strsplit(value.dptr, "\n", 0);
	*messages = NULL;
	for (i = 0; strv[i] != NULL; i++) {
		*messages = g_list_prepend(*messages, g_strdup(strv[i]));
	}
	g_strfreev(strv);
	*messages = g_list_reverse(*messages);

	free(value.dptr);

	/* now delete those messages */
	gdbm_delete(info->msgdb, key);
	g_free(id_lower);

	return TRUE;
}

static gboolean
client_gdbm_add_message (void *blob, const char *id, const char *message)
{
	GDBMInfo *info = (GDBMInfo *)blob;
	datum key, value;
	char *id_lower;
	char *newval;
	int ret;

	id_lower = g_strdup(id);
	g_strdown(id_lower);
	key.dptr = id_lower;
	key.dsize = strlen(id_lower)+1;
	value = gdbm_fetch(info->msgdb, key);

	newval = g_strdup_printf("%s%s%s", (value.dptr == NULL) ? "" : (char *)value.dptr,
				 (value.dptr == NULL) ? "" : "\n",
				 message);
	if (value.dptr != NULL) {
		free(value.dptr);
	}
	value.dptr = newval;
	value.dsize = strlen(newval)+1;
	ret = gdbm_store(info->msgdb, key, value, GDBM_REPLACE);
	g_free(newval);
	g_free(id_lower);

	return (ret == 0);
}

/* ----------  buddy list stuff  ---------- */

static char **
get_buddy_list (GDBMInfo *info, const char *id, gboolean reverse)
{
        char *id_lower;
        datum key, value;
        char **strv;

        id_lower = g_strdup_printf("%c%s", (reverse ? '<' : '>'), id);
        g_strdown(id_lower);
        key.dptr = id_lower;
        key.dsize = strlen(id_lower)+1;
        value = gdbm_fetch(info->buddydb, key);

        strv = NULL;
        if (value.dptr != NULL) {
                if (strlen((char *)value.dptr) != 0) {
                        strv = g_strsplit((char *)value.dptr, ",", 0);
                }
                free(value.dptr);
        }
        g_free(id_lower);

        return strv;
}

static gboolean
change_buddy_list (GDBMInfo *info, const char *id, gboolean reverse, char **strv,
                   const char *add, const char *remove)
{
        char *id_lower;
        GString *gstr = g_string_new("");
        datum key, value;
        int i, ret;
        gboolean done_first = FALSE;

        if (strv != NULL) {
                for (i = 0; strv[i] != NULL; i++) {
                        if ((remove == NULL) || (g_strcasecmp(strv[i], remove) != 0)) {
                                if (done_first) {
                                        gstr = g_string_append_c(gstr, ',');
                                }
                                done_first = TRUE;
                                gstr = g_string_append(gstr, strv[i]);
                        }
                }
        }
        if (add != NULL) {
                if (done_first) {
                        gstr = g_string_append_c(gstr, ',');
                }
                gstr = g_string_append(gstr, add);
        }

        id_lower = g_strdup_printf("%c%s", (reverse ? '<' : '>'), id);
        g_strdown(id_lower);
        key.dptr = id_lower;
        key.dsize = strlen(id_lower)+1;
        value.dptr = gstr->str;
        value.dsize = strlen(gstr->str)+1;
        ret = gdbm_store(info->buddydb, key, value, GDBM_REPLACE);
        g_string_free(gstr, TRUE);
        g_free(id_lower);

        return (ret == 0) ? TRUE : FALSE;
}

static gboolean
get_buddy_info (GDBMInfo *info, const char *id, const char *buddy, BuddyFlags *flags, char **message, int *denycount)
{
        char *id_lower;
        datum key, value;
        EnamlBlock *block;
        const char *flagstr;

        id_lower = g_strdup_printf("%s/%s", id, buddy);
        g_strdown(id_lower);
        key.dptr = id_lower;
        key.dsize = strlen(id_lower)+1;
        value = gdbm_fetch(info->buddydb, key);

        *flags = 0;
        *message = NULL;
        *denycount = 0;

        if (value.dptr == NULL) {
                g_free(id_lower);
                return FALSE;
        }
        block = enaml_parse(value.dptr);
        flagstr = enaml_get_string(block, "flags", "");
        if (strchr(flagstr, 'B') != NULL) {
                *flags = BUDDY_APPROVED | BUDDY_BLOCKED;
        } else if (strchr(flagstr, 'A') != NULL) {
                *flags = BUDDY_APPROVED;
        } else if (strchr(flagstr, 'P') != NULL) {
                *flags = 0;
        } else if (strchr(flagstr, 'D') != NULL) {
                *flags = BUDDY_DENIED;
        }
        if (strchr(flagstr, 'S') != NULL) {
                *flags |= BUDDY_STORED;
        }
        *message = g_strdup(enaml_get_string(block, "message", NULL));
        *denycount = (int)enaml_get_long(block, "denycount", 0);
        
        free(value.dptr);
        g_free(id_lower);
        enaml_free(block);
        return TRUE;
}

static gboolean
put_buddy_info (GDBMInfo *info, const char *id, const char *buddy, BuddyFlags flags, const char *message, 
                int denycount)
{
        char *id_lower;
        char *dump;
        datum key, value;
        EnamlBlock *block;
        char flagstr[10];
        int ret;

        if (flags & BUDDY_BLOCKED) {
                flagstr[0] = 'B';
        } else if (flags & BUDDY_APPROVED) {
                flagstr[0] = 'A';
        } else if (flags & BUDDY_DENIED) {
                flagstr[0] = 'D';
        } else {
                flagstr[0] = 'P';
        }
        if (flags & BUDDY_STORED) {
                flagstr[1] = 'S';
                flagstr[2] = '\0';
        } else {
                flagstr[1] = '\0';
        }

        block = enaml_new();
        enaml_add_string(block, "flags", flagstr, 0);
        if (message != NULL) {
                enaml_add_string(block, "message", message, 0);
        }
        enaml_add_long(block, "denycount", (long)denycount);
        dump = enaml_compose(block, FALSE);

        id_lower = g_strdup_printf("%s/%s", id, buddy);
        g_strdown(id_lower);
        key.dptr = id_lower;
        key.dsize = strlen(id_lower)+1;
        value.dptr = dump;
        value.dsize = strlen(dump)+1;
	ret = gdbm_store(info->buddydb, key, value, GDBM_REPLACE);
        g_free(dump);
        enaml_free(block);
        g_free(id_lower);

        return (ret == 0) ? TRUE : FALSE;
}

static gboolean
client_gdbm_get_buddies (void *blob, Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        char **strv;
        BuddyFlags flags;
        char *message;
        int denycount;
        int i;

        strv = get_buddy_list(info, client->id, FALSE);
        if (strv != NULL) {
                for (i = 0; strv[i] != NULL; i++) {
                        get_buddy_info(info, client->id, strv[i], &flags, &message, &denycount);
                        client_buddy_add(client, strv[i], flags, message, denycount);
                        g_free(message);
                }
                g_strfreev(strv);
        }

        strv = get_buddy_list(info, client->id, TRUE);
        if (strv != NULL) {
                for (i = 0; strv[i] != NULL; i++) {
                        get_buddy_info(info, strv[i], client->id, &flags, &message, &denycount);
                        client_buddy_add_reverse(client, strv[i], flags, message);
                        g_free(message);
                }
                g_strfreev(strv);
        }

        return TRUE;
}

static gboolean
client_gdbm_update_buddy (void *blob, const Client *client, const Buddy *buddy)
{
        GDBMInfo *info = (GDBMInfo *)blob;

        return put_buddy_info(info, client->id, buddy->id, buddy->flags, buddy->message, buddy->deny_count);
}

static gboolean
client_gdbm_add_buddy (void *blob, const Client *client, const Buddy *buddy)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        char **strv;

        if (put_buddy_info(info, client->id, buddy->id, buddy->flags, buddy->message, buddy->deny_count)) {
                strv = get_buddy_list(info, client->id, FALSE);
                change_buddy_list(info, client->id, FALSE, strv, buddy->id, NULL);
                if (strv != NULL) {
                        g_strfreev(strv);
                }

                strv = get_buddy_list(info, buddy->id, TRUE);
                change_buddy_list(info, buddy->id, TRUE, strv, client->id, NULL);
                if (strv != NULL) {
                        g_strfreev(strv);
                }
                return TRUE;
        }

        return FALSE;
}

static gboolean
client_gdbm_delete_buddy (void *blob, const Client *client, const Buddy *buddy)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        char **strv;
        char *id_lower;
        datum key;

        id_lower = g_strdup_printf("%s/%s", client->id, buddy->id);
        g_strdown(id_lower);
        key.dptr = id_lower;
        key.dsize = strlen(id_lower)+1;
        if (gdbm_delete(info->buddydb, key) == 0) {
                strv = get_buddy_list(info, client->id, FALSE);
                change_buddy_list(info, client->id, FALSE, strv, NULL, buddy->id);
                if (strv != NULL) {
                        g_strfreev(strv);
                }

                strv = get_buddy_list(info, buddy->id, TRUE);
                change_buddy_list(info, buddy->id, TRUE, strv, NULL, client->id);
                if (strv != NULL) {
                        g_strfreev(strv);
                }
                g_free(id_lower);
                return TRUE;
        }
        g_free(id_lower);
        return FALSE;
}

DBModule *
client_gdbm_get_module (void)
{
        DBModule *module;

        module = g_new0 (DBModule, 1);
        module->init = &client_gdbm_init;
        module->shutdown = &client_gdbm_shutdown;
        module->add_client = &client_gdbm_add_client;
        module->update_login = &client_gdbm_update_login;
        module->update_logout = &client_gdbm_update_logout;
        module->update_client = &client_gdbm_update_client;
        module->lookup_client = &client_gdbm_lookup_client;
        module->check_password = &client_gdbm_check_password;
        module->set_password = &client_gdbm_set_password;
        module->get_seen = &client_gdbm_get_seen;
	module->get_messages = &client_gdbm_get_messages;
	module->add_message = &client_gdbm_add_message;
        module->get_buddies = &client_gdbm_get_buddies;
        module->update_buddy = &client_gdbm_update_buddy;
        module->add_buddy = &client_gdbm_add_buddy;
        module->delete_buddy = &client_gdbm_delete_buddy;
        module->empty = NULL;

        return module;
}
