/* $Id: client-sql.c,v 1.24 2001/08/20 01:04:24 robey Exp $
 *
 * Do SQL queries to fetch/create users/channels, filling in Client entries.
 *
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <say2/say2.h>
#include "client.h"
#include "client-db.h"
#include "client-sql.h"


static void *
client_sql_init (const GData *params)
{
        const char *host, *user, *pass, *db;

        host = (const char *) g_datalist_get_data ((GData **)&params, "host");
        user = (const char *) g_datalist_get_data ((GData **)&params, "username");
        pass = (const char *) g_datalist_get_data ((GData **)&params, "password");
        db = (const char *) g_datalist_get_data ((GData **)&params, "database");

        return (void *) sql_init (host, user, pass, db);
}

static void
client_sql_shutdown (void *blob)
{
        SQLState *state = (SQLState *)blob;

        sql_shutdown (state);
}

static gboolean
client_sql_add_client (void *blob, const Client *client)
{
        SQLState *state = (SQLState *)blob;
        GString *values;
        char *email, *tagline, *privs, *prefs, *icon;
        SQLError err;
        int rows;

        values = g_string_new ("insert into channel ");
        if (client->flags & CLIENT_CHANNEL) {
                g_string_append (values, "(id, email, tagline, prefs, created, icon)");
        } else {
                g_string_append (values, "(id, email, tagline, prefs, created, icon, privs)");
        }
        g_string_append (values, " values (");

        g_string_sprintfa (values, "'%s',", client->id);

        email = sql_quote_string (client->email);
        g_string_sprintfa (values, "%s,", email);
        g_free (email);

        tagline = sql_quote_string (client->tagline);
        g_string_sprintfa (values, "%s,", tagline);
        g_free (tagline);

        prefs = client_prefs_to_string (client);
        g_string_sprintfa (values, "'%s',", prefs);
        g_free (prefs);

        g_string_sprintfa (values, "NOW(),");

        icon = sql_quote_buffer (client->icon, client->iconsize);
        g_string_sprintfa (values, "%s", icon);
        g_free (icon);

        if (! (client->flags & CLIENT_CHANNEL)) {
                privs = client_privs_to_string (client);
                g_string_sprintfa (values, ",'%s',", privs);
                g_free (privs);
        }

        err = sql_update (state, values->str, &rows);
        g_string_free (values, TRUE);

        return (err == SQL_OK ? TRUE : FALSE);
}

static gboolean
client_sql_update_login (void *blob, const Client *client)
{
        SQLState *state = (SQLState *)blob;
        char *quoted_id, *query;
        char *table_name;
        SQLError err;
        int rows;

        if (client->flags & CLIENT_CHANNEL) {
                table_name = "channel";
        } else {
                table_name = "user";
        }
        quoted_id = sql_quote_string (client->id);
        query = g_strconcat ("update ", table_name, " set last_login=NOW() where id=", quoted_id, NULL);
        g_free (quoted_id);

        err = sql_update (state, query, &rows);
        g_free (query);
        return (err == SQL_OK);
}

static gboolean
client_sql_update_logout (void *blob, const Client *client)
{
        SQLState *state = (SQLState *)blob;
        char *quoted_id, *query;
        char *table_name;
        SQLError err;
        int rows;

        if (client->flags & CLIENT_CHANNEL) {
                table_name = "channel";
        } else {
                table_name = "user";
        }
        quoted_id = sql_quote_string (client->id);
        query = g_strconcat ("update ", table_name, " set last_logout=NOW() where id=", quoted_id, NULL);
        g_free (quoted_id);

        err = sql_update (state, query, &rows);
        g_free (query);
        return (err == SQL_OK);
}

/* writes new email/tagline/prefs/privs to the sql server */
static gboolean
client_sql_update_client (void *blob, const Client *client)
{
        SQLState *state = (SQLState *)blob;
        char *quoted_id, *quoted_email, *quoted_tagline, *quoted_icon, *quoted_attrs;
        char *privs, *prefs, *attrs;
        char *query;
        char *privs_tag, *table_tag;
        SQLError err;
        int rows;

        quoted_id = sql_quote_string (client->id);
        quoted_email = sql_quote_string (client->email);
        quoted_tagline = sql_quote_string (client->tagline);
        quoted_icon = sql_quote_buffer (client->icon, client->iconsize);
        prefs = client_prefs_to_string (client);
        if (enaml_block_size (client->attributes) > 0) {
                attrs = enaml_compose (client->attributes, FALSE);
        } else {
                attrs = NULL;
        }
        quoted_attrs = sql_quote_string (attrs);
        if (client->flags & CLIENT_CHANNEL) {
                table_tag = "channel";
                privs = g_strdup ("");
        } else {
                table_tag = "user";
                privs_tag = client_privs_to_string (client);
                privs = g_strdup_printf (",privs='%s'", privs_tag);
                g_free (privs_tag);
        }
        query = g_strconcat ("update ", table_tag, " set email=", quoted_email, ",tagline=",
                             quoted_tagline, privs, ",prefs='", prefs, "',icon=", quoted_icon,
                             ",attributes=", quoted_attrs,
                             " where id=", quoted_id, NULL);
        g_free (quoted_id);
        g_free (quoted_email);
        g_free (quoted_tagline);
        g_free (quoted_icon);
        g_free (privs);
        g_free (prefs);
        g_free (quoted_attrs);
        enaml_compose_free (attrs);

    log("query: %s", query);
        err = sql_update (state, query, &rows);
        g_free (query);

        /* even if rows=0, might be ok: might not have changed anything */
        if (err != SQL_OK) {
                return FALSE;
        } else {
                return TRUE;
        }
}

static gboolean
client_sql_lookup_client (void *blob, const char *id, Client *client)
{
        SQLState *state = (SQLState *)blob;
        char *quoted_id;
        char *query;
        char *attrs;
        const char *table_name, *tags;
        SQLError err;
        GList *list;
        char **row;

        if ((client != NULL) && (client->flags & CLIENT_CHANNEL)) {
                table_name = "channel";
                tags = "id, email, tagline, prefs, icon, attributes";
        } else {
                table_name = "user";
                tags = "id, email, tagline, prefs, icon, attributes, privs";
        }

        quoted_id = sql_quote_string (id);
        query = g_strconcat ("select ", tags, " from ", table_name, " where id=", quoted_id, NULL);
        g_free (quoted_id);

        err = sql_query (state, query, &list);
        g_free (query);

        if (err != SQL_OK) {
                return FALSE;
        }

        if (g_list_length (list) == 0) {
                return FALSE;
        }

        if (g_list_length (list) != 1) {
                log ("SQL: somehow a client lookup matched %d times", g_list_length (list));
                sql_free_list (list);
                return FALSE;
        }

        if (client != NULL) {
                row = (char **)(list->data);

                g_free (client->id);
                client->id = sql_unquote_string (row[0]);
                if (! (client->flags & CLIENT_CHANNEL)) {
                        client_string_to_privs (client, row[6]);
                }
                client_string_to_prefs (client, row[3]);
                g_free (client->nick);
                client->nick = g_strdup (client->id);
                g_free (client->email);
                client->email = sql_unquote_string (row[1]);
                g_free (client->tagline);
                if (row[2] != NULL) {
                        client->tagline = sql_unquote_string (row[2]);
                        if (strlen (client->tagline) == 0) {
                                g_free (client->tagline);
                                client->tagline = NULL;
                        }
                } else {
                        client->tagline = NULL;
                }
                if (row[4] != NULL) {
                        client->icon = sql_unquote_buffer (row[4], &client->iconsize);
                }
                if (row[5] != NULL) {
                        attrs = sql_unquote_string (row[5]);
                        enaml_free (client->attributes);
                        client->attributes = enaml_parse (attrs);
                        g_free (attrs);
                }
        }

        sql_free_list (list);
        return TRUE;
}

static gboolean
client_sql_check_password (void *blob, const char *id, const char *password, int flags)
{
        SQLState *state = (SQLState *)blob;
        char *query;
        char *table_name;
        char *quoted_id, *quoted_password;
        SQLError err;
        GList *out;
        gboolean result;

        if (flags & CLIENT_CHANNEL) {
                table_name = "channel";
        } else {
                table_name = "user";
        }

        quoted_id = sql_quote_string (id);
        quoted_password = sql_quote_string (password);
        query = g_strconcat ("select id from ", table_name, " where id=", quoted_id,
                             " and password=PASSWORD(", quoted_password, ")", NULL);
        g_free (quoted_id);
        g_free (quoted_password);

        err = sql_query (state, query, &out);
        g_free (query);
        if (err != SQL_OK) {
                return FALSE;
        }

        result = (g_list_length (out) > 0);
        sql_free_list (out);
        return result;
}

static gboolean
client_sql_set_password (void *blob, const Client *client, const char *password)
{
        SQLState *state = (SQLState *)blob;
        char *query;
        char *table_name;
        char *quoted_id, *quoted_password;
        SQLError err;
        int rows;

        if (client->flags & CLIENT_CHANNEL) {
                table_name = "channel";
        } else {
                table_name = "user";
        }

        quoted_id = sql_quote_string (client->id);
        quoted_password = sql_quote_string (password);
        query = g_strconcat ("update ", table_name, " set password=PASSWORD(", quoted_password,
                             ") where id=", quoted_id, NULL);
        g_free (quoted_id);
        g_free (quoted_password);

        err = sql_update (state, query, &rows);
        g_free (query);
        if ((err != SQL_OK) || (rows != 1)) {
                return FALSE;
        } else {
                return TRUE;
        }
}

static gboolean
client_sql_get_seen (void *blob, const char *id, time_t *last_login, time_t *last_logout, int flags)
{
        SQLState *state = (SQLState *)blob;
        SQLError err;
        GList *list;
        const char *table_name;
        char *quoted_id, *query;
        char **row;

        if (flags & CLIENT_CHANNEL) {
                table_name = "channel";
        } else {
                table_name = "user";
        }

        quoted_id = sql_quote_string (id);
        query = g_strconcat ("select UNIX_TIMESTAMP(last_login),UNIX_TIMESTAMP(last_logout) from ",
                             table_name, " where id=",
                             quoted_id, NULL);
        g_free (quoted_id);

        err = sql_query (state, query, &list);
        g_free (query);

        if (err != SQL_OK) {
                return FALSE;
        }
        if (g_list_length (list) == 0) {
                return FALSE;
        }
        if (g_list_length (list) != 1) {
                log ("SQL: somehow a client lookup matched %d times", g_list_length (list));
                sql_free_list (list);
                return FALSE;
        }

        row = (char **)(list->data);
        *last_login = (time_t) atol (row[0]);
        *last_logout = (time_t) atol (row[1]);
        sql_free_list (list);
        return TRUE;
}

static gboolean
client_sql_get_messages (void *blob, const Client *client, GList **messages)
{
	SQLState *state = (SQLState *)blob;
	SQLError err;
	GList *list, *iter;
	char *query, **row;
	int rows;

	query = g_strconcat("select message from message where dest='", client->id, "'", NULL);
	err = sql_query(state, query, &list);
	g_free(query);

	if (err != SQL_OK) {
		return FALSE;
	}
	*messages = NULL;
	for (iter = g_list_first(list); iter != NULL; iter = g_list_next(iter)) {
		row = (char **)(iter->data);
		*messages = g_list_prepend(*messages, g_strdup(row[0]));
	}
	*messages = g_list_reverse(*messages);

	/* now, delete them all */
	if (g_list_length(list) > 0) {
		query = g_strconcat("delete from message where dest='", client->id, "'", NULL);
		err = sql_update(state, query, &rows);
		/* error is non-fatal (the messages will just get echoed again next time) */
		g_free(query);
	}
        sql_free_list(list);

	return TRUE;
}

static gboolean
client_sql_add_message (void *blob, const char *id, const char *message)
{
	SQLState *state = (SQLState *)blob;
	SQLError err;
	char *quoted_text, *query;
	int rows;

	quoted_text = sql_quote_string(message);
	query = g_strconcat("insert into message (dest, message) values ('", id, "',",
			    quoted_text, ")", NULL);
	err = sql_update(state, query, &rows);
	g_free(query);
	g_free(quoted_text);

	return (err == SQL_OK ? TRUE : FALSE);
}

static BuddyFlags
buddy_flags_from_sql (const char *str)
{
        BuddyFlags flags = 0;

        switch(str[0]) {
        case 'b':
        case 'B':
                /* blocked */
                flags = BUDDY_APPROVED | BUDDY_BLOCKED;
                break;
        case 'a':
        case 'A':
                /* approved */
                flags = BUDDY_APPROVED;
                break;
        case 'p':
        case 'P':
                /* pending */
                flags = 0;
                break;
        case 'd':
        case 'D':
                /* denied */
                flags = BUDDY_DENIED;
                break;
        }
        if (str[1] == 'S') {
                flags |= BUDDY_STORED;
        }
        return flags;
}

static void
buddy_flags_to_sql (BuddyFlags flags, char *str)
{
        if (flags & BUDDY_DENIED) {
                str[0] = 'D';
        } else if (flags & BUDDY_APPROVED) {
                if (flags & BUDDY_BLOCKED) {
                        str[0] = 'B';
                } else {
                        str[0] = 'A';
                }
        } else {
                str[0] = 'P';
        }
        if (flags & BUDDY_STORED) {
                str[1] = 'S';
                str[2] = '\0';
        } else {
                str[1] = '\0';
        }
}

static gboolean
client_sql_get_buddies (void *blob, Client *client)
{
        SQLState *state = (SQLState *)blob;
        SQLError err;
        char *query, **row;
        GList *list, *iter;
        BuddyFlags flags;

        query = g_strconcat("select buddy,status,message,denycount from buddy where id='",
                            client->id, "'", NULL);
        err = sql_query(state, query, &list);
        g_free(query);

        if (err != SQL_OK) {
                return FALSE;
        }
        for (iter = g_list_first(list); iter != NULL; iter = g_list_next(iter)) {
                row = (char **)iter->data;
                flags = buddy_flags_from_sql(row[1]);
                client_buddy_add(client, row[0], flags, row[2], atoi(row[3]));
        }
        sql_free_list(list);

        query = g_strconcat("select id,status,message from buddy where buddy='", client->id, "'", NULL);
        err = sql_query(state, query, &list);
        g_free(query);

        if (err != SQL_OK) {
                return FALSE;
        }
        for (iter = g_list_first(list); iter != NULL; iter = g_list_next(iter)) {
                row = (char **)iter->data;
                flags = buddy_flags_from_sql(row[1]);
                client_buddy_add_reverse(client, row[0], flags, row[2]);
        }
        sql_free_list(list);

        return TRUE;
}

static gboolean
client_sql_update_buddy (void *blob, const Client *client, const Buddy *buddy)
{
        SQLState *state = (SQLState *)blob;
	SQLError err;
        char *quoted_message, *query;
        char status_str[10], deny_count_str[10];
        int rows;

        buddy_flags_to_sql(buddy->flags, status_str);
        sprintf(deny_count_str, "%d", buddy->deny_count);
        quoted_message = sql_quote_string(buddy->message);
        query = g_strconcat("update buddy set status='", status_str, "',message=", quoted_message,
                            ",denycount=", deny_count_str, " where id='", client->id, "' and buddy='", buddy->id,
                            "'", NULL);
        err = sql_update(state, query, &rows);
        g_free(query);
        g_free(quoted_message);

        return ((err == SQL_OK) && (rows > 0) ? TRUE : FALSE);
}

static gboolean
client_sql_add_buddy (void *blob, const Client *client, const Buddy *buddy)
{
        SQLState *state = (SQLState *)blob;
	SQLError err;
        char *quoted_message, *query;
        char status_str[10], deny_count_str[10];
        int rows;

        buddy_flags_to_sql(buddy->flags, status_str);
        sprintf(deny_count_str, "%d", buddy->deny_count);
        quoted_message = sql_quote_string(buddy->message);
        query = g_strconcat("insert into buddy (id,buddy,status,message,denycount) values ('", client->id, "','",
                            buddy->id, "','", status_str, "',", quoted_message, ",", deny_count_str, ")", NULL);
        err = sql_update(state, query, &rows);
        g_free(query);
        g_free(quoted_message);

        return ((err == SQL_OK) && (rows > 0) ? TRUE : FALSE);
}

static gboolean
client_sql_delete_buddy (void *blob, const Client *client, const Buddy *buddy)
{
        SQLState *state = (SQLState *)blob;
        SQLError err;
        char *query;
        int rows;

        query = g_strconcat("delete from buddy where id='", client->id, "' and buddy='", buddy->id, "'", NULL);
        err = sql_update(state, query, &rows);
        g_free(query);

        return ((err == SQL_OK) && (rows > 0) ? TRUE : FALSE);
}

DBModule *
client_sql_get_module (void)
{
        DBModule *module;

        module = g_new0 (DBModule, 1);
        module->init = &client_sql_init;
        module->shutdown = &client_sql_shutdown;
        module->add_client = &client_sql_add_client;
        module->update_login = &client_sql_update_login;
        module->update_logout = &client_sql_update_logout;
        module->update_client = &client_sql_update_client;
        module->lookup_client = &client_sql_lookup_client;
        module->check_password = &client_sql_check_password;
        module->set_password = &client_sql_set_password;
        module->get_seen = &client_sql_get_seen;
	module->get_messages = &client_sql_get_messages;
	module->add_message = &client_sql_add_message;
        module->get_buddies = &client_sql_get_buddies;
        module->update_buddy = &client_sql_update_buddy;
        module->add_buddy = &client_sql_add_buddy;
        module->delete_buddy = &client_sql_delete_buddy;
        module->empty = NULL;

        return module;
}
