/* $Id: client.c,v 1.35 2001/11/04 01:07:03 robey Exp $
 *
 * manipulates client structs.
 * FIXME: these should be called "sessions" now.
 *
 * Copyright (C) 2000, 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <glib.h>
#include <say2/say2.h>
#include "client.h"
#include "client-db.h"

/* idle for more than this many seconds, and you're officially "idle" */
#define IDLE_LIMIT	600

static GList *client_list = NULL;	/* GList of Client */
static long client_counter = 1;


Client *
client_new (gboolean add_to_roster)
{
        Client *client;

        client = g_new0 (Client, 1);
        client->magic = CLIENT_MAGIC;
        client->id = g_strdup_printf ("(%ld)", client_counter++);
        client->flags = CLIENT_USER | CLIENT_NOTIFY;
        client->login = client->activity = client->status_changed = time(NULL);

        if (add_to_roster) {
                client_list = g_list_append (client_list, client);
        }

        return client;
}

void
client_new_connection (Client *client, Socket *sock, const char *host, guint16 port)
{
        ClientConnection *conn;

        conn = g_new0(ClientConnection, 1);
	conn->socket = sock;
	conn->host = g_strdup(host);
	conn->port = port;
        client->conn_list = g_list_prepend(client->conn_list, conn);

        /* link the client and socket */
        if (sock != NULL) {
                socket_set_data (sock, client);
        }
}

static void
client_conn_free (ClientConnection *conn)
{
	g_free(conn->host);
	conn->host = NULL;
        g_free(conn->version);
        conn->version = NULL;
	conn->port = 0;
	conn->socket = NULL;
	g_free(conn);
}

static void
client_buddy_free (Buddy *buddy)
{
        g_free(buddy->id);
        g_free(buddy->message);
        g_free(buddy);
}

/* doesn't remove the user data from the socket */
void
client_free (Client *client)
{
        g_assert (IS_CLIENT (client));

        client->magic = 0;
	g_list_foreach(client->conn_list, (GFunc)client_conn_free, NULL);
	g_list_free(client->conn_list);
        g_list_foreach(client->buddy_list, (GFunc)client_buddy_free, NULL);
        g_list_free(client->buddy_list);
        g_list_foreach(client->reverse_buddy_list, (GFunc)client_buddy_free, NULL);
        g_list_free(client->reverse_buddy_list);
        g_free (client->id);
        g_free (client->nick);
        g_free (client->email);
        g_free (client->tagline);
        g_free (client->status);
        g_free (client->address);
        enaml_free (client->attributes);

        client_list = g_list_remove (client_list, client);

        g_free (client);
}

/* drop a reference to a client:
 * if the client is online, do nothing.  otherwise free the memory.
 */
void
client_discard (Client *client)
{
        if (client->flags & CLIENT_OFFLINE) {
                client_free(client);
        }
}

Buddy *
client_buddy_add (Client *client, const char *id, BuddyFlags flags, const char *message, int deny_count)
{
        Buddy *buddy;

        buddy = g_new0(Buddy, 1);
        buddy->id = g_strdup(id);
        buddy->flags = flags;
        buddy->message = g_strndup(message, CLIENT_MAX_BUDDY_MESSAGE);
        buddy->deny_count = deny_count;
        client->buddy_list = g_list_prepend(client->buddy_list, buddy);
        return buddy;
}

Buddy *
client_buddy_add_reverse (Client *client, const char *id, BuddyFlags flags, const char *message)
{
        Buddy *buddy;

        buddy = g_new0(Buddy, 1);
        buddy->id = g_strdup(id);
        buddy->flags = flags;
        buddy->message = g_strndup(message, CLIENT_MAX_BUDDY_MESSAGE);
        client->reverse_buddy_list = g_list_prepend(client->reverse_buddy_list, buddy);
        return buddy;
}

Buddy *
client_buddy_find (Client *client, const char *id)
{
        GList *iter;
        Buddy *buddy;

        for (iter = g_list_first(client->buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if (g_strcasecmp(id, buddy->id) == 0) {
                        return buddy;
                }
        }
        return NULL;
}

Buddy *
client_buddy_find_reverse (Client *client, const char *id)
{
        GList *iter;
        Buddy *buddy;

        for (iter = g_list_first(client->reverse_buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if (g_strcasecmp(id, buddy->id) == 0) {
                        return buddy;
                }
        }
        return NULL;
}

gboolean
client_buddy_remove (Client *client, const char *id)
{
        Buddy *buddy;

        buddy = client_buddy_find(client, id);
        if (buddy == NULL) {
                return FALSE;
        }
        client->buddy_list = g_list_remove(client->buddy_list, buddy);
        client_buddy_free(buddy);
        return TRUE;
}

gboolean
client_buddy_remove_reverse (Client *client, const char *id)
{
        Buddy *buddy;

        buddy = client_buddy_find_reverse(client, id);
        if (buddy == NULL) {
                return FALSE;
        }
        client->reverse_buddy_list = g_list_remove(client->reverse_buddy_list, buddy);
        client_buddy_free(buddy);
        return TRUE;
}

void
client_show_buddy_status (Client *client)
{
        GList *iter;
        Buddy *buddy;
        Client *buddy_client;
        EnamlBlock *block;

        for (iter = g_list_first(client->buddy_list); iter != NULL; iter = g_list_next(iter)) {
                buddy = (Buddy *)iter->data;
                if (buddy->flags & BUDDY_APPROVED) {
                        buddy_client = client_by_id(buddy->id, CLIENT_USER);
                        if ((buddy_client != NULL) && !(buddy->flags & BUDDY_BLOCKED)) {
                                block = client_make_status(buddy_client, TRUE);
                                client_send(client, block);
                                enaml_free(block);
                        } else {
                                buddy_client = client_new(FALSE);
                                buddy_client->nick = g_strdup(buddy->id);
                                buddy_client->id = g_strdup(buddy->id);
                                block = client_make_status(buddy_client, FALSE);
                                client_send(client, block);
                                enaml_free(block);
                                client_free(buddy_client);
                        }
                }
        }
}

static ClientConnection *
client_conn_by_socket (Client *client, const Socket *sock)
{
	GList *iter;
	ClientConnection *conn;

	for (iter = g_list_first(client->conn_list); iter != NULL; iter = g_list_next(iter)) {
		conn = (ClientConnection *)iter->data;
		if (conn->socket == sock) {
			return conn;
		}
	}
	return NULL;
}

int
client_count (void)
{
        return g_list_length (client_list);
}

gboolean
client_move_socket (Client *client, Client *from_client, const Socket *sock)
{
	ClientConnection *conn;

	if (g_list_length(client->conn_list) >= CLIENT_MAX_SOCKETS) {
		return FALSE;
	}

	conn = client_conn_by_socket(from_client, sock);
	if (conn == NULL) {
		return FALSE;
	}
	client->conn_list = g_list_prepend(client->conn_list, conn);
	from_client->conn_list = g_list_remove(from_client->conn_list, conn);
	return TRUE;
}

gboolean
client_remove_socket (Client *client, const Socket *sock)
{
	ClientConnection *conn;

	conn = client_conn_by_socket(client, sock);
	if (conn != NULL) {
		client_conn_free(conn);
		client->conn_list = g_list_remove(client->conn_list, conn);
		return TRUE;
	}
	return FALSE;
}

void
client_close_all_sockets_but (Client *client, const Socket *sock)
{
	GList *iter;
	ClientConnection *conn;

        iter = g_list_first(client->conn_list);
        while (iter != NULL) {
		conn = (ClientConnection *)iter->data;
		if (conn->socket == sock) {
			/* save this one */
                        iter = g_list_next(iter);
		} else {
			socket_close(conn->socket);
                        iter = g_list_first(client->conn_list);
		}
	}
}

void
client_close_all_sockets (Client *client)
{
	client_close_all_sockets_but (client, NULL);
}

gboolean
client_has_sockets (Client *client)
{
        return (g_list_length(client->conn_list) > 0);
}

const char *
client_host_by_socket (Client *client, const Socket *sock)
{
	ClientConnection *conn;

	conn = client_conn_by_socket(client, sock);
	if (conn != NULL) {
		return conn->host;
	} else {
		return NULL;
	}
}

const char *
client_get_version (const Client *client)
{
        GList *iter;
        ClientConnection *conn;

        for (iter = g_list_first(client->conn_list); iter != NULL; iter = g_list_next(iter)) {
                conn = (ClientConnection *)iter->data;
                if (conn->version != NULL) {
                        return conn->version;
                }
        }
        return NULL;
}

void
client_set_version (Client *client, char *version)
{
        ClientConnection *conn;

        if ((client->conn_list == NULL) || (client->conn_list->data == NULL)) {
                return;
        }
        conn = (ClientConnection *)client->conn_list->data;
        conn->version = version;
}

Client *
client_by_nick (const char *nick)
{
        GList *iter;
        Client *client;

        for (iter = g_list_first (client_list); iter; iter = g_list_next (iter)) {
                client = (Client *)(iter->data);
                g_assert (IS_CLIENT (client));
                if (client->nick && (g_strcasecmp (client->nick, nick) == 0)) {
                        return client;
                }
        }

        return NULL;
}

Client *
client_by_id (const char *id, ClientFlags flags)
{
        GList *iter;
        Client *client;

        for (iter = g_list_first (client_list); iter; iter = g_list_next (iter)) {
                client = (Client *)(iter->data);
                g_assert (IS_CLIENT (client));
                if (client->nick && (g_strcasecmp (client->id, id) == 0) &&
                    ((client->flags & flags) == flags)) {
                        return client;
                }
        }

        return NULL;
}

static int 
grep_string (const char *grep, const char *orig)
{
        int i;

	if (strlen (grep) > strlen (orig)) {
		return 0;
	}
        for (i = 0; i <= strlen (orig) - strlen (grep); i++) {
                if (strncasecmp (orig+i, grep, strlen (grep)) == 0) {
                        return 1;
                }
        }
        return 0;
}

GList *
client_grep (const char *grep)
{
        GList *iter;
        GList *out = NULL;
        Client *client;

        for (iter = g_list_first (client_list); iter != NULL; iter = g_list_next (iter)) {
                client = (Client *)(iter->data);
                g_assert (IS_CLIENT (client));
                if (client->nick) {
                        if (!grep || grep_string (grep, client->nick)) {
                                out = g_list_append (out, client);
                        }
                }
        }
        return out;
}

/* put the list of client nicknames in an enaml block */
static void
client_get_list_int (EnamlBlock *block, const char *key, const char *grep, int ids, ClientFlags flags)
{
        GList *iter;
        Client *client;
        int count = 0;
        EnamlList *list;

        list = enaml_new_list ();
        for (iter = g_list_first (client_list); iter; iter = g_list_next (iter)) {
                client = (Client *)(iter->data);
                g_assert (IS_CLIENT (client));
                if (client->nick) {
                        if ((client->flags & flags) != flags) {
                                continue;
                        }
                        if (!grep || (grep_string (grep, client->nick))) {
                                count++;
                                if (ids) {
                                        enaml_list_add_string (list, client->id, 0);
                                } else {
                                        enaml_list_add_string (list, client->nick, 0);
                                }
                        }
                }
        }
        enaml_add_list (block, key, list, ENAML_FREE);
}

void
client_get_list (EnamlBlock *block, const char *key, const char *grep, ClientFlags flags)
{
        client_get_list_int (block, key, grep, 0, flags);
}

void
client_get_id_list (EnamlBlock *block, const char *key, const char *grep, ClientFlags flags)
{
        client_get_list_int (block, key, grep, 1, flags);
}

void
client_notify_signon (Client *client)
{
        GList *iter;
        Client *notify;
        EnamlBlock *out;

        g_assert (IS_CLIENT (client));

        out = enaml_buildf ("signon", "nick:%s id:%s email:%s time:%d ",
                            client->nick, client->id,
                            client->email, (long)client->login);

        for (iter = g_list_first (client_list); iter; iter = g_list_next (iter)) {
                notify = (Client *)(iter->data);
                g_assert (IS_CLIENT (notify));
                if ((notify->flags & CLIENT_NOTIFY) && (notify != client) &&
		    notify->nick) {
                        /* tell this client about the new signon */
                        client_send(notify, out);
                }
        }
        enaml_free (out);
}

void
client_notify_signoff (Client *client)
{
        GList *iter;
        Client *notify;
        EnamlBlock *out;

        out = enaml_buildf ("signoff", "nick:%s id:%s time:%d", client->nick,
                            client->id, (long) time (NULL));

        for (iter = g_list_first (client_list); iter; iter = g_list_next (iter)) {
                notify = (Client *)(iter->data);
                g_assert (IS_CLIENT (notify));
                if ((notify->flags & CLIENT_NOTIFY) && (notify != client) &&
                    client->nick && notify->nick) {
                        /* tell this client about the new signoff */
                        client_send(notify, out);
                }
        }
        enaml_free (out);
}

void
client_send_text (const Client *client, const char *text)
{
	GList *iter;
	ClientConnection *conn;

	for (iter = g_list_first(client->conn_list); iter != NULL; iter = g_list_next(iter)) {
		conn = (ClientConnection *)iter->data;
		if (conn->socket != NULL) {
                        socket_write(conn->socket, text, strlen(text));
                        socket_write(conn->socket, "\r\n", 2);
		}
	}
}

void
client_send (const Client *client, const EnamlBlock *block)
{
	GList *iter;
	ClientConnection *conn;

	for (iter = g_list_first(client->conn_list); iter != NULL; iter = g_list_next(iter)) {
		conn = (ClientConnection *)iter->data;
		if (conn->socket != NULL) {
                        send_enaml_block(conn->socket, block);
		}
	}
}

void
client_sendf (const Client *client, const char *tag, const char *fmt, ...)
{
        EnamlBlock *out;
        va_list arg;

        va_start(arg, fmt);
        out = enaml_vbuildf(tag, fmt, arg);
        va_end(arg);

        client_send(client, out);
        enaml_free(out);
}

void
client_broadcast (const EnamlBlock *block)
{
        GList *iter;
        Client *client;

        for (iter = g_list_first (client_list); iter; iter = g_list_next (iter)) {
                client = (Client *)(iter->data);
                g_assert (IS_CLIENT (client));
                if (client->nick) {
                        client_send(client, block);
                }
        }
}

void
client_got_ping (Client *client, Socket *sock)
{
	ClientConnection *conn;

	conn = client_conn_by_socket(client, sock);
	if (conn != NULL) {
		conn->ping_pending = FALSE;
	} else {
		log ("BAD BAD BAD ping from unknown socket: %s", client->id);
	}
}

/* kill all clients that aren't answering pings */
void
client_ping_all (void)
{
        GList *iter, *iter2;
        GList *doom = NULL;
        Client *client;
	ClientConnection *conn;
	Socket *sock;
        time_t now;

        now = time (NULL);
        for (iter = g_list_first (client_list); iter; iter = g_list_next (iter)) {
                client = (Client *)(iter->data);
		for (iter2 = g_list_first(client->conn_list); iter2 != NULL;
		     iter2 = g_list_next(iter2)) {
			conn = (ClientConnection *)iter2->data;
			if (conn->ping_pending) {
                                log ("%s[%s:%d] ping timeout", client->id, conn->host, conn->port);
                                send_enaml_buildf(conn->socket, "error", "text:%s", "Ping timeout.");
                                doom = g_list_append (doom, conn->socket);
                        } else {
                                send_enaml_buildf(conn->socket, "ping", NULL);
				conn->ping_pending = TRUE;
                        }
                }
        }

        for (iter = g_list_first (doom); doom; doom = g_list_next (doom)) {
                sock = (Socket *)(doom->data);
                socket_close(sock);
        }
        g_list_free (doom);
}                

/* lookup an offline client record */
Client *
client_lookup_offline (void *db_state, const char *id, int flags)
{
        Client *client;

        client = client_new (FALSE);
        client->flags = flags;
        if (! db_lookup_client (db_state, id, client)) {
                /* no such client */
                client_free (client);
                return NULL;
        }

        client->login = (time_t)0L;
        client->flags |= CLIENT_OFFLINE;
        return client;
}

/* this is meant to be called after client_get_target, and with the same error block.
 * (it may reuse the error message from before, or it may generate a new one.)
 */
Client *
client_get_offline_target (void *db_state, const EnamlBlock *block, ClientFlags flags,
			   EnamlBlock **error_block)
{
	const char *id = NULL;
	Client *client = NULL;

	id = enaml_get_string(block, "id", NULL);
	if (id == NULL) {
		/* use same error block */
		return NULL;
	}
	client = client_lookup_offline(db_state, id, flags);
	if (client == NULL) {
		enaml_free(*error_block);
		*error_block = enaml_buildf("no-such-client", "id:%s type:%s", id,
					    (flags & CLIENT_USER ? "user" :
					     (flags & CLIENT_CHANNEL ? "channel" : "any")));
		return NULL;
	}

	client->login = (time_t)0L;
	client->flags |= CLIENT_OFFLINE;
	return client;
}

/* return the Client* that is the target of an enaml command
 * (indexes either the "nick" or "id").
 * flags should be CLIENT_USER OR CLIENT_CHANNEL, depending on what you're looking for.
 * if it returns NULL, error_block will be filled in with a response
 */
Client *
client_get_target (const EnamlBlock *block, ClientFlags flags, EnamlBlock **error_block)
{
        const char *id = NULL;
        const char *nick = NULL;
        Client *client;

        *error_block = NULL;

        id = enaml_get_string (block, "id", NULL);
        if (id != NULL) {
                client = client_by_id (id, flags);
        } else {
                nick = enaml_get_string (block, "nick", NULL);
                if (nick != NULL) {
                        flags &= CLIENT_CHANNEL;
                        flags |= CLIENT_USER;
                        client = client_by_nick (nick);
                } else {
                        /* didn't supply an id or nick */
                        *error_block = enaml_buildf ("error", "text:%s",
                                                     "Command requires an id or nick.");
                        return NULL;
                }
        }

        if (!client || !client->nick) {
                if (id) {
                        *error_block = enaml_buildf ("not-present", "id:%s type:%s", id,
                                                     (flags & CLIENT_USER ? "user" :
                                                      (flags & CLIENT_CHANNEL ? "channel" : "any")));
                } else {
                        *error_block = enaml_buildf ("not-present", "nick:%s type:%s", nick,
                                                     (flags & CLIENT_USER ? "user" :
                                                      (flags & CLIENT_CHANNEL ? "channel" : "any")));
                }
        }

        return client;
}

const char *
client_mode_string (const Client *client)
{
        g_assert (IS_CLIENT (client));

        switch (client->mode) {
        case CLIENT_MODE_ACTIVE:
                if (time (NULL) - client->activity > IDLE_LIMIT) {
                        return "idle";
                }
                return "active";
        case CLIENT_MODE_BUSY:
                return "busy";
        case CLIENT_MODE_AWAY:
                return "away";
        default:
                return "???";
        }
}

EnamlBlock *
client_make_status (const Client *client, gboolean online)
{
        EnamlBlock *block;

        if (online) {
                block = enaml_buildf ("status", "nick:%s id:%s messages:%s time:%d mode:%s text:%s",
                                      client->nick, client->id,
                                      (client->flags & CLIENT_NO_MSG) ? "off" : "on",
                                      (long)time(NULL),
                                      client_mode_string (client), client->status);
        } else {
                block = enaml_buildf("status", "nick:%s id:%s time:%d mode:%s",
                                     client->nick, client->id, (long)time(NULL), "offline");
        }
        return block;
}

char *
client_privs_to_string (const Client *client)
{
        char *out;
        int i;

        g_assert (IS_CLIENT (client));

        out = g_malloc (10);
        i = 0;
        if (client->privs & PRIV_ADDUSER) {
                out[i++] = 'A';
        }
	if (client->privs & PRIV_BOOTUSER) {
		out[i++] = 'B';
	}
        if (client->privs & PRIV_DEBUG) {
                out[i++] = 'D';
        }
        if (client->privs & PRIV_WALL) {
                out[i++] = 'W';
        }
        out[i] = 0;
        return out;
}

char *
client_prefs_to_string (const Client *client)
{
        char *out;
        int i;

        g_assert (IS_CLIENT (client));

        out = g_malloc (10);
        i = 0;
        if (client->flags & CLIENT_NOTIFY) {
                out[i++] = 'N';
        }
        out[i] = 0;
        return out;
}

void
client_string_to_privs (Client *client, const char *str)
{
        const char *p;

        g_assert (IS_CLIENT (client));

        client->privs = 0;
        if (str == NULL) {
                return;
        }
        for (p = str; *p; p++) {
                switch (*p) {
                case 'A':
                        client->privs |= PRIV_ADDUSER;
                        break;
                case 'B':
			client->privs |= PRIV_BOOTUSER;
                        break;
                case 'D':
                        client->privs |= PRIV_DEBUG;
                        break;
                case 'W':
                        client->privs |= PRIV_WALL;
                        break;
                }
        }
}

void
client_string_to_prefs (Client *client, const char *str)
{
        const char *p;

        g_assert (IS_CLIENT (client));

        client->flags &= (CLIENT_USER|CLIENT_CHANNEL);
        if (str == NULL) {
                return;
        }
        for (p = str; *p; p++) {
                if (*p == 'N') {
                        client->flags |= CLIENT_NOTIFY;
                }
        }
}

int
client_type_to_flags (const char *type)
{
        if (strcasecmp (type, "user") == 0) {
                return CLIENT_USER;
        } else if (strcasecmp (type, "channel") == 0) {
                return CLIENT_CHANNEL;
        } else {
                return 0;
        }
}

/* this is for debugging */
char *
client_to_string (Client *client, gboolean verbose)
{
        GString *out;
        char *prefstr, *privstr;
        char *outstr;
	GList *iter;
	ClientConnection *conn;
        Buddy *buddy;

        g_assert (IS_CLIENT (client));

        out = g_string_new ("");
        g_string_sprintfa (out, "<client '%s' nick[%s]", client->id, client->nick);
        prefstr = client_prefs_to_string (client);
        privstr = client_privs_to_string (client);
        g_string_sprintfa (out, " prefs[%s] privs[%s]", prefstr, privstr);
        if (client->flags & CLIENT_USER) {
                g_string_sprintfa (out, " user");
        } else if (client->flags & CLIENT_CHANNEL) {
                g_string_sprintfa (out, " channel");
        } else {
                g_string_sprintfa (out, " ?(%X)", client->flags);
        }
        g_free (prefstr);
        g_free (privstr);
        if (verbose) {
		if (client->address != NULL) {
			g_string_sprintfa(out, " address[%s]", client->address);
		}
                g_string_sprintfa (out, " email[%s] tagline[%s]", client->email,
                                   client->tagline != NULL ? client->tagline : "");
                g_string_sprintfa (out, " mode[%s] status[%s]", client_mode_string (client),
                                   client->status != NULL ? client->status : "");
                g_string_sprintfa (out, " login=%lu activity=%lu", (unsigned long)client->login,
                                   (unsigned long)client->activity);
		for (iter = g_list_first(client->conn_list); iter != NULL;
		     iter = g_list_next(iter)) {
			conn = (ClientConnection *)iter->data;
			g_string_sprintfa(out, " {sock=%08X, host[%s:%d] client[%s]}",
					  (unsigned int)conn->socket,
					  conn->host, (int)conn->port,
                                          conn->version != NULL ? conn->version : "");
		}

                /* buddy list */
                g_string_sprintfa(out, " buddies[");
                for (iter = g_list_first(client->buddy_list); iter != NULL; iter = g_list_next(iter)) {
                        buddy = (Buddy *)iter->data;
                        if (iter != g_list_first(client->buddy_list)) {
                                g_string_sprintfa(out, " ");
                        }
                        g_string_sprintfa(out, "%s", buddy->id);
                        if (!(buddy->flags & (BUDDY_APPROVED|BUDDY_DENIED))) {
                                g_string_sprintfa(out, "(P)");
                        }
                        if (buddy->flags & BUDDY_BLOCKED) {
                                g_string_sprintfa(out, "(B)");
                        }
                        if (buddy->flags & BUDDY_DENIED) {
                                g_string_sprintfa(out, "(D:%d)", buddy->deny_count);
                        }
                }
                g_string_sprintfa(out, "]");
                                          
        }
        g_string_sprintfa (out, ">");

        outstr = out->str;
        g_string_free (out, FALSE);

        return outstr;
}
