/* $Id: memory.c,v 1.5 2001/07/04 20:54:46 robey Exp $
 *
 * A memory is logged or state data, connected to a channel, that
 * clients can modify.  (Might be useful for implementing a whiteboard,
 * chess game, etc).
 *
 * Copyright (C) 2000, 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <say2/say2.h>
#include "client.h"
#include "memory.h"


typedef enum {
        MEMORY_LOG = 0,
        MEMORY_STATE
} MemoryType;

typedef struct {
        char *name;
        MemoryType type;
        GList *log;		/* (list of [EnamlBlock]) */
        GHashTable *state;     	/* (key -> value [EnamlBlock]) */
} Memory;


/* memories in existence: (key -> Memory) */
GHashTable *memories = NULL;


static void
send_error (Socket *sock, const char *command, const char *msg)
{
        send_enaml_buildf (sock, "error", "command:%s text:%s", command, msg);
}

void
memories_init (void)
{
        if (memories != NULL) {
                /* already init */
                return;
        }

        memories = g_hash_table_new (g_str_hash, g_str_equal);
}

static gboolean
memory_new (const char *name, MemoryType type)
{
        char *small_name;
        Memory *memory;

        small_name = g_strdup (name);
        g_strdown (small_name);
        if (g_hash_table_lookup (memories, small_name) != NULL) {
                /* already exists */
                g_free (small_name);
                return FALSE;
        }

        memory = g_new0 (Memory, 1);
        memory->name = small_name;
        memory->type = type;
        switch (type) {
        case MEMORY_LOG:
                memory->log = NULL;
                break;
        case MEMORY_STATE:
                memory->state = g_hash_table_new (g_str_hash, g_str_equal);
                break;
        }

        g_hash_table_insert (memories, small_name, memory);
        return TRUE;
}

static gboolean
memory_clear_callback (char *key, EnamlBlock *value, void *unused)
{
        g_free (key);
        enaml_free (value);
        return TRUE;
}

static gboolean
memory_clear (const char *name)
{
        char *small_name;
        Memory *memory;
        GList *iter;
        EnamlBlock *block;

        small_name = g_strdup (name);
        g_strdown (small_name);
        memory = (Memory *) g_hash_table_lookup (memories, small_name);
        g_free (small_name);
        if (! memory) {
                return FALSE;
        }

        if (memory->type == MEMORY_LOG) {
                /* clear out the log */
                for (iter = g_list_first (memory->log); iter; iter = g_list_next (iter)) {
                        block = (EnamlBlock *) (iter->data);
                        enaml_free (block);
                }
                g_list_free (memory->log);
                memory->log = NULL;
        } else {
                /* clear out the hashtable */
                g_hash_table_foreach_remove (memory->state, (GHRFunc)memory_clear_callback, NULL);
        }
        return TRUE;
}


/* replace a hashtable entry in a state-based memory */
static void
set_memory_entry (Memory *memory, const EnamlBlock *block)
{
        const char *key, *str_value;
        char *key_small, *key_orig;
        const EnamlBlock *value;
        EnamlBlock *copy, *old_block;

        key = enaml_get_string (block, "key", "default");
        key_small = g_strdup (key);
        g_strdown (key_small);
        value = enaml_get_block (block, "value");
        if (value) {
                copy = enaml_copy_block (value);
        } else {
                /* if they made a string instead, stick it in a fake
                 * block for them.
                 */
                str_value = enaml_get_string (block, "value", NULL);
                if (str_value) {
                        copy = enaml_new();
                        enaml_add_string (copy, "value", str_value, ENAML_COPY);
                } else {
                        copy = NULL;
                }
        }

        /* erase old entry, if there is one */
        if (g_hash_table_lookup_extended (memory->state, key_small, (void **)&key_orig,
                                          (void **)&old_block)) {
                g_hash_table_remove (memory->state, key_orig);
                g_free (key_orig);
                enaml_free (old_block);
        }

        if (copy) {
                g_hash_table_insert (memory->state, key_small, copy);
        }
}

/* check for memory items in a say block */
void
memories_scan (const EnamlBlock *block)
{
        EnamlIter iter;
        const char *key;
        char *key_small;
        Memory *memory;
        EnamlBlock *copy;
        const EnamlBlock *sub_block;

        /* unlikely that there will be more entries in the enaml block than the number
         * of memories (since most messages will have exactly one block), so we iterate
         * the keys of the enaml block.
         */
        for (key = enaml_iter_first (block, &iter); key; key = enaml_iter_next (&iter)) {
                if (enaml_type (block, key) != ENAML_KEY_BLOCK) {
                        continue;
                }

                sub_block = enaml_get_block (block, key);
                if (sub_block == NULL) {
                        continue;
                }

                key_small = g_strdup (key);
                g_strdown (key_small);
                memory = (Memory *) g_hash_table_lookup (memories, key_small);
                if (memory) {
                        if (memory->type == MEMORY_LOG) {
                                /* add the whole block to the log */
                                copy = enaml_copy_block (sub_block);
                                memory->log = g_list_append (memory->log, copy);
                        } else {
                                /* replace an entry in the hash */
                                set_memory_entry (memory, sub_block);
                        }
                }
                g_free (key_small);
        }
}

struct PassBlock {
        Client *client;
        int so_far;
        int max;
};

/* for state-based memory */
static void
dump_memory_callback (const char *key, const EnamlBlock *block, struct PassBlock *pb)
{
        pb->so_far++;
        if (pb->so_far > pb->max) {
                return;
        }
        client_sendf(pb->client, "dump-memory", "key:%s value:%!B",
                     key, block);
}

/* dump a memory's contents to a client */
static gboolean
memory_dump (const char *name, Client *client, int max)
{
        Memory *memory;
        char *small_name;
        GList *iter;
        long count, total, i;
        struct PassBlock pb;

        small_name = g_strdup (name);
        g_strdown (small_name);
        memory = (Memory *) g_hash_table_lookup (memories, small_name);
        g_free (small_name);
        if (! memory) {
                return FALSE;
        }

        if (memory->type == MEMORY_LOG) {
                total = g_list_length (memory->log);
        } else {
                total = g_hash_table_size (memory->state);
        }
        count = total;
        if ((max > 0) && (count > max)) {
                count = max;
        }

        client_sendf(client, "begin-response", "count:%d total:%d response:%s",
                     count, total, "dump-memory");
        if (memory->type == MEMORY_LOG) {
                i = 0;
                for (iter = g_list_first (memory->log); iter != NULL; iter = g_list_next (iter)) {
                        i++;
                        if (i <= total-count) {
                                continue;
                        }
                        client_sendf (client, "dump-memory", "%s:%!B",
                                      memory->name, (EnamlBlock *)(iter->data));
                }
        } else {
                pb.so_far = 0;
                pb.max = count;
                pb.client = client;
                g_hash_table_foreach (memory->state, (GHFunc)dump_memory_callback, &pb);
        }

        client_sendf(client, "end-response", "response:%s", "dump-memory");
        return TRUE;
}

static void
memories_list_callback (const char *key, const Memory *memory, EnamlBlock *block)
{
        EnamlList *list;

        if (memory->type == MEMORY_LOG) {
                list = enaml_get_list (block, "log");
                if (! list) {
                        list = enaml_new_list ();
                        enaml_add_list (block, "log", list, ENAML_FREE);
                }
                enaml_list_add_string (list, memory->name, 0);
        } else {
                /* state */
                list = enaml_get_list (block, "state");
                if (! list) {
                        list = enaml_new_list ();
                        enaml_add_list (block, "state", list, ENAML_FREE);
                }
                enaml_list_add_string (list, memory->name, 0);
        }
}

static void
memories_list (Client *client)
{
        EnamlBlock *block, *memblock;

        block = enaml_new ();
        g_hash_table_foreach (memories, (GHFunc)memories_list_callback, block);

        memblock = enaml_new();
        enaml_add_block (memblock, "memories", block, ENAML_FREE);
        client_send(client, memblock);
        enaml_free (memblock);
}

void
do_list_memories (Client *client, const EnamlBlock *block, Socket *sock)
{
        memories_list (client);
}

void
do_new_memory (Client *client, const EnamlBlock *block, Socket *sock)
{
        const char *type_str;
        const char *name;
        MemoryType type;

        if (block == NULL) {
                send_error (sock, "new-memory", "Required new-memory field: name");
                return;
        }

        type_str = enaml_get_string (block, "type", "state");
        name = enaml_get_string (block, "name", NULL);

        if (name == NULL) {
                send_error (sock, "new-memory", "Required new-memory field: name");
                return;
        }
        if (strcasecmp (type_str, "state") == 0) {
                type = MEMORY_STATE;
        } else if (strcasecmp (type_str, "log") == 0) {
                type = MEMORY_LOG;
        } else {
                send_error (sock, "new-memory", "Type must be log or state");
                return;
        }

        if (! memory_new (name, type)) {
                send_error (sock, "new-memory", "Memory already exists");
                return;
        }

        /* on success, dump a list of existing memories */
        memories_list (client);
}

void
do_clear_memory (Client *client, const EnamlBlock *block, Socket *sock)
{
        const char *name;

        if (block == NULL) {
                send_error (sock, "clear-memory", "Required clear-memory field: name");
                return;
        }

        name = enaml_get_string (block, "name", NULL);
        if (name == NULL) {
                send_error (sock, "clear-memory", "Required clear-memory field: name");
                return;
        }

        if (! memory_clear (name)) {
                send_error (sock, "clear-memory", "Memory doesn't exist");
                return;
        }

        client_sendf(client, "clear-memory", "name:%s", name);
}

void
do_dump_memory (Client *client, const EnamlBlock *block, Socket *sock)
{
        const char *name;
        long max;

        if (block == NULL) {
                send_error (sock, "dump-memory", "Required dump-memory field: name");
                return;
        }

        name = enaml_get_string (block, "name", NULL);
        if (name == NULL) {
                send_error (sock, "dump-memory", "Required dump-memory field: name");
                return;
        }

        max = enaml_get_long (block, "max", 0);
        if (! memory_dump (name, client, (int)max)) {
                send_error (sock, "dump-memory", "Memory doesn't exist");
                return;
        }
}
